static char rcsid[] = "@(#)$Id: shared_mailer.c,v 2.9 2016/03/21 20:26:14 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.9 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *                  (was hurtta+elm@posti.FMI.FI, hurtta+elm@ozone.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

/*
 * This file is compiled only if dlopen() is available, so
 * that file does not need to be guarded with #ifdef 
 */

#include "def_mailer.h"

DEBUG_VAR(Debug,__FILE__,"dl");

#include "shared_imp.h"
#include "mailer_imp.h"

#include "rc_imp.h"
#include "save_opts.h"


#if ANSI_C
#define S_(x) static x;
#else
#define S_(x)
#endif

S_(verify_mail_domain_f NO_verify_mail_domain)
static enum verify_domain_status 
  NO_verify_mail_domain P_((const char *domain,char **rewrite,
			    int translated_value  /* -1 if not share specific */,
			    const struct dt_enum_shared * is_shared_value,
			    size_t dt_enum_shared_size
			    ));
static enum verify_domain_status NO_verify_mail_domain (domain,rewrite,
							translated_value,
							is_shared_value,
							dt_enum_shared_size)
     const char *domain;
     char **rewrite;
     int translated_value  /* -1 if not share specific */;
     const struct dt_enum_shared * is_shared_value;
     size_t dt_enum_shared_size;
{
    return vd_status_no_routine;
}


#define SHARED_MAILER_magic	0xF904

struct sl_mailer_data {
    uint16            magic;      

    verify_mail_domain_f * verify_mdomain;
};


S_(sl_reg_functions sl_reg_mailer)
static int sl_reg_mailer P_((struct ImpInfo *i,
			      int reg_idx));
static int sl_reg_mailer(i,reg_idx)
     struct ImpInfo *i;
     int reg_idx;
{
    /* union hack to avoid warnings about casting
     * between pointer-to-object and pointer-to-function
     */
    
    int res = 0;
    
    union F14 {
	void *                       ptr;
	provides_shared_MCF_f      * f14;
    } f14;
    
    union F18 {
	void *                       ptr;
	verify_mail_domain_f       * f18;
    } f18;

    struct reg_list *r = & (i->regs[reg_idx]);

     if (&use_shared_mailer != r->var)
	panic("SHARED PANIC",__FILE__,__LINE__,
	      "sl_reg_mailer",
	      "Wrong variable",0);

     if (r->r.mailer->magic != SHARED_MAILER_magic)
	 panic("SHARED PANIC",__FILE__,__LINE__,
	       "sl_reg_mailer",
	       "Bad magic bumber",0);
     
    f14.ptr = dlsym(i->handle, "provides_shared_MCF");
    f18.ptr = dlsym(i->handle, "verify_mail_domain");

    if (!f14.f14) {
	DPRINT(Debug,7,(&Debug, " ... NO provides_shared_MCF\n"));
    }
    
    if (!f18.f18) {
	DPRINT(Debug,7,(&Debug, " ... NO verify_mail_domain\n"));
    }
    

    if (f14.f14) {
	int count;
	size_t s_res14           = 0; 
	size_t s_opt_size        = 0;
	size_t s_send_state_size = 0;
	size_t s_cancel_size     = 0;
	size_t s_mailer_info_size = 0;
	size_t s_env_from_size   = 0;
	
	struct mailer_config *res14 = f14.f14(&count,&s_res14,
					      &s_opt_size,
					      &s_send_state_size,
					      &s_cancel_size,
					      &s_mailer_info_size,
					      &s_env_from_size);
	
	if (s_res14 != sizeof (*res14)) {
	    DPRINT(Debug,1,(&Debug,
			    "...  struct mailer_config mismatch: %d should be %d\n",
			    s_res14,sizeof (*res14)));
	} else if (s_opt_size != sizeof (struct mailer_option_list)) {
	    DPRINT(Debug,1,(&Debug,
			    "...  struct mailer_option_list mismatch: %d should be %d\n",
			    s_opt_size,sizeof (struct mailer_option_list)));

	} else if (s_send_state_size != sizeof (struct mail_send_state)) {
	    DPRINT(Debug,1,(&Debug,
			    "...  struct  mail_send_state mismatch: %d should be %d\n",
			    s_send_state_size,
			    sizeof (struct mail_send_state)));

	} else if (s_cancel_size != sizeof (struct mailer_cancel_ticket)) {
	    DPRINT(Debug,1,(&Debug,
			    "...  struct mailer_cancel_ticket mismatch: %d should be %d\n",
			    s_cancel_size,
			    sizeof (struct mailer_cancel_ticket)));
		   
	} else if (s_mailer_info_size != sizeof (struct mailer_info)) {
	    DPRINT(Debug,1,(&Debug,
			    "...  struct mailer_mailer_info mismatch: %d should be %d\n",
			    
			    s_mailer_info_size,sizeof (struct mailer_info)));

	} else if (s_env_from_size != sizeof (struct mailer_env_from)) {

	    DPRINT(Debug,1,(&Debug,
			    "...  struct mailer_env_from mismatch: %d should be %d\n",
			    s_env_from_size,sizeof (struct mailer_env_from)));


	} else {		
	    int x;
	    
	    shared_MCF_types = 
		safe_array_realloc(shared_MCF_types,
				   (shared_MCF_type_count + count),
				   sizeof (shared_MCF_types[0]));
	    
	    DPRINT(Debug,7,(&Debug,
			    " ... provides_shared_MCF:  (shared_MCF_types) count %d\n",
			    count));
	    
	    for (x = 0; x < count; x++) {
		
		if (MAILER_CONFIG_magic != res14[x].magic) {
		    DPRINT(Debug,1,(&Debug, 
				    "  ...  struct mailer_config [%d] magic mismatch %d (should be %d)\n",
				    x,res14[x].magic,MAILER_CONFIG_magic));
		    goto failure;
		}
				
		shared_MCF_types[shared_MCF_type_count+x].T = &(res14[x]);
		shared_MCF_types[shared_MCF_type_count+x].imp_idx = i;
	    }
	    shared_MCF_type_count += count;
	    res = 1;
	}

    }

    if (f18.f18) {
	DPRINT(Debug,7,(&Debug, " ... verify_mail_domain\n"));
	
	r->r.mailer->verify_mdomain = f18.f18;
	res = 1;
    }
    
 failure:
    i->regs[reg_idx].valid   = res;

    DPRINT(Debug,7,(&Debug, 
		    "sl_reg_mailer:  [%p]->regs[%d].valid = %d\n",
		    i,reg_idx,res));

    
    return res;
}

S_(sl_zero_reg_list sl_zero_mailer)
static void sl_zero_mailer P_((struct dt_shared_info *var,
				struct reg_list *r));
static void sl_zero_mailer(var,r)
     struct dt_shared_info *var;
     struct reg_list *r;
{
    r->var     = var;
    r->valid   = 0;
    r->r.mailer  = safe_malloc(sizeof (*r->r.mailer));   /* never freed ... */
    
    r->r.mailer->verify_mdomain = NO_verify_mail_domain;
    r->r.mailer->magic = SHARED_MAILER_magic;
}

S_(sl_unreg_functions sl_unreg_mailer)
static void sl_unreg_mailer P_((struct ImpInfo *i, int reg_idx));
static void sl_unreg_mailer(i,reg_idx)
     struct ImpInfo *i; 
     int reg_idx;
{
    struct reg_list *r = & (i->regs[reg_idx] );
    r->valid   = 0;

}

static struct shared_loader mailer_loader = {
    SHARED_LOADER_magic,
    sl_reg_mailer,
    sl_zero_mailer,
    sl_unreg_mailer
};

SHAREDLIB use_shared_mailer = {
    &mailer_loader            /* loader */,
    0,                       
    NULL, 0,

    NULL
};

struct mailer_config  * loc_mailer_type(f)
     char *f;
{
    int x;
    
    /* We must load all mailer libraries */
    load_shared_libs1(&use_shared_mailer); 

    for (x = 0; x < shared_MCF_type_count; x++) {
	if (0 == strcmp(shared_MCF_types[x].T->mailer_type,f)) {
	  
	  DPRINT(Debug,10,(&Debug, 
			   "loc_mailer_type=%p: f=%s, shared_MCF_types[%d]\n",
			   shared_MCF_types[x].T,f,x));
	  
	  return shared_MCF_types[x].T;
	}
    }

    DPRINT(Debug,10,(&Debug, 
		     "loc_mailer_type=NULL: f=%s, shared_MCF_type_count=%d\n",
		     f,shared_MCF_type_count));

    return NULL;
}

static enum verify_domain_status process_one_verify_domain 
    P_((const char * domain,
	char      ** rewrite,
	int translated_value,
	const struct dt_enum_shared * is_shared_value,
	int var_index));
static enum verify_domain_status process_one_verify_domain(domain,rewrite,
							   translated_value,
							   is_shared_value,
							   var_index)
     const char * domain;
     char      ** rewrite;
     int translated_value;
     const struct dt_enum_shared * is_shared_value;
     int var_index;
{
    if (var_index < 0 || var_index >= use_shared_mailer.shared_list_len)
	    panic("SHARED PANIC",__FILE__,__LINE__,
		  "process_one_verify_domain",
		  "Bad var index",0);

    if (use_shared_mailer.shared_list[var_index]->handle) {
	int r = give_rnum(use_shared_mailer.shared_list[var_index],
			  &use_shared_mailer);

	struct sl_mailer_data * mailer = 
	    use_shared_mailer.shared_list[var_index]->regs[r].r.mailer;

	if (SHARED_MAILER_magic != mailer->magic)
	    panic("SHARED PANIC",__FILE__,__LINE__,
		  "process_one_verify_domain",
		  "Bad magic number",0);

	if (use_shared_mailer.shared_list[var_index]->regs[r].valid)
	    return mailer->verify_mdomain(domain,
					  rewrite,
					  translated_value,
					  is_shared_value,
					  sizeof (* is_shared_value));
    } 
	
    return vd_status_no_routine;
}

enum verify_domain_result shared_verify_domain(domain,rewrite,tag,
					       translated_value,
					       is_shared_value)
     const char * domain;
     char      ** rewrite;
     const char * tag; 
     int          translated_value;
     const struct dt_enum_shared * is_shared_value;
{
    enum verify_domain_status r = vd_status_no_routine;

    if (tag) {
	int basic_index = locate_and_load_library1(tag,
						   &use_shared_mailer);

	if (basic_index < 0) {
	    DPRINT(Debug,9,(&Debug, 
			    "shared_verify_domain=%d verify_domain_not_available: %s, tag %s not found\n",
			    verify_domain_not_available,domain,tag));

	    return verify_domain_not_available;
	}

	r = process_one_verify_domain(domain,rewrite,translated_value,
				      is_shared_value,basic_index);
	
    } else { 
	enum verify_domain_status saved_r = vd_status_no_routine;
	int i;
	
	/* must load all libraries */
	load_shared_libs1(&use_shared_mailer);
	
	for (i = 0; i < use_shared_mailer.shared_list_len; i++) {
	    
	    r = process_one_verify_domain(domain,rewrite,-1,NULL,i);
	    
	    if (vd_status_fallback == r)
		saved_r = r;
	    else if (vd_status_no_routine != r)
		break;

	    if (rewrite && *rewrite) {
		free(*rewrite);
		*rewrite = NULL;
	    }
	}
	
	if (vd_status_no_routine == r)
	    r = saved_r;
    }

    switch(r) {
    case vd_status_no_routine: 
	DPRINT(Debug,9,(&Debug, 
			"shared_verify_domain=%d verify_domain_not_available: %s\n",
			verify_domain_not_available,domain));
	return verify_domain_not_available;
    case vd_status_fallback:   
	DPRINT(Debug,9,(&Debug, 
			"shared_verify_domain=%d verify_domain_fallback: %s\n",
			verify_domain_fallback,domain));
	return verify_domain_fallback;
    case vd_status_failure:    
	DPRINT(Debug,9,(&Debug, 
			"shared_verify_domain=%d verify_domain_failure: %s\n",
			verify_domain_failure,domain));
	return verify_domain_failure;
    case vd_status_not_found:  
	DPRINT(Debug,9,(&Debug, 
			"shared_verify_domain=%d verify_domain_not_found: %s\n",
			verify_domain_not_found,domain));
	return verify_domain_not_found;
    case vd_status_verified:    
	DPRINT(Debug,9,(&Debug, 
			"shared_verify_domain=%d verify_domain_ok: %s",
			verify_domain_ok,domain));

	if (rewrite && *rewrite) {
	    DPRINT(Debug,9,(&Debug,", rewrite to: %s",
			    *rewrite));
	}
	DPRINT(Debug,9,(&Debug,"\n"));
	return verify_domain_ok;
    case vd_domain_no_mail:
	DPRINT(Debug,9,(&Debug, 
			"shared_verify_domain=%d verify_domain_no_mail: %s",
			verify_domain_no_mail,domain));

	if (rewrite && *rewrite) {
	    DPRINT(Debug,9,(&Debug,", rewrite to: %s",
			    *rewrite));
	}
	DPRINT(Debug,9,(&Debug,"\n"));
	return verify_domain_no_mail;
    case vd_domain_bad_syntax:
	DPRINT(Debug,9,(&Debug, 
			"shared_verify_domain=%d verify_domain_bad_syntax: %s\n",
			verify_domain_bad_syntax,domain));
	return verify_domain_bad_syntax;
    }

    return verify_domain_failure;

}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
