/* $Id: def_mbox.h,v 2.15 2017/05/03 19:34:22 hurtta Exp $ */

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.15 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> (was hurtta+elm@posti.FMI.FI)
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/

#include "elm_defs.h"
#include "connection_imp.h"

#include "addrlib.h"
#include "mboxlib.h"
#include "mbx_imp.h"

#include "mime.h"


#ifdef USE_DLOPEN
extern SHAREDLIB use_shared_mbox;
#endif

extern struct out_state_type state_out_dir;

extern void remote_server_dec_dir P_((struct folder_browser *browser));
extern void remote_server_inc_dir P_((struct folder_browser *browser,
				      const struct remote_server *server));

extern struct remote_server  * const LOCAL_SERVER;   /* static, not alloced dynamically */

extern struct remote_server_type * const POP_SERVER_TYPE;
extern struct remote_server_type * const IMAP_SERVER_TYPE;

extern void remove_remote_server_from_folder P_((struct folder_info   * folder,
						 struct remote_server * server));
extern struct remote_server * give_remote_server_from_folder P_((struct folder_info        * folder,
								 struct remote_server_type * t,
								 struct browser_passhm     * passhm));


extern void remove_remote_server_from_connection P_((struct connection_cache * con,
						     struct remote_server    * server));
extern struct remote_server * give_remote_server_from_connection P_((struct connection_cache   * con,
								     struct remote_server_type * t,
								     struct browser_passhm     * passhm));


/* remote_server.c */

extern struct browser_type remote_browser;

extern void remote_browser_failure P_((struct folder_browser *dir)); 
extern int change_remote_browser_server P_((struct folder_browser *dir));

extern struct remote_server * remote_browser_server P_((struct folder_browser *dir));

extern struct browser_type
 * remote_server_real_type P_((struct remote_server *server));

extern int connect_remote_server_to_browser P_((
   struct remote_server *server,
   struct folder_browser *dir));


/* savefolder.c */

extern struct browser_type dummy_browser;

extern void dummy_browser_set_remote P_((struct folder_browser *dir,
					 const char *remote /* may be NULL */,
					 struct browser_passhm *passhm));

#if 0
extern const char * dummy_browser_get_remote P_((struct folder_browser *dir));
#endif


/* 0 == new type is same than old
 * 1 == type is changed
 */
extern int change_folder_browser_type P_((struct folder_browser *dir,
					  struct browser_type *new_type));

extern void malloc_keep_browser_write_state P_((WRITE_STATE * ptr));
extern void free_keep_browser_write_state P_((WRITE_STATE * ptr));

/* pop.c */

extern struct remote_account *set_pop_connection P_((struct folder_info    *fh,
						     struct remote_account *X,
						     struct browser_passhm *passhm ));

extern const struct remote_account * give_pop_remote_account P_((struct folder_info *fh));

/* increments refcount of passhm */
extern void pop_mbx_set_passhm P_((struct folder_info *fh,
				   struct browser_passhm * passhm));


/* connection.c */

extern const struct remote_account * give_con_remote_account P_((struct connection_cache *c));

/* imap.c */

extern struct remote_account *set_imap_connection P_((struct folder_info    *fh,
						      struct remote_account *X,
						      struct browser_passhm *passhm));

struct remote_account *set_imap_connection_C P_((struct folder_info *fh,
						 struct connection_cache * CX,
						 struct browser_passhm *passhm));

extern struct connection_cache *give_imap_connection P_((struct folder_browser *dir));

/* increments refcount of passhm */
extern void imap_browser_set_passhm P_((struct folder_browser *dir,
					struct browser_passhm  * passhm));

/* increments refcount of passhm */
extern void imap_mbx_set_passhm P_((struct folder_info *fh,
				   struct browser_passhm * passhm));

extern int imap_same_host_user_nopath P_((const struct folder_info *fh,
					  const struct remote_account *ra2));

/* savefolder.c */

extern int change_dir_helper P_((struct folder_browser *dir,
				 struct string **buffer));

extern int valid_browser P_((struct browser_type    * type));

extern int select_dir_item_helper P_((struct folder_browser *dir,
				      struct string **buffer,
				      int *newpos));

/* mbox_generic.c */


#if ANSI_C
#define E_(x) extern x;
#else
#define E_(x)
#endif

/* Returns 1 if changed,
   -1 error
 */
E_(browser_reset_filter_dir browser_reset_filter_generic)
extern int browser_reset_filter_generic P_((struct folder_browser *dir));


E_(browser_select_idx_dir browser_select_idx_generic)
extern int browser_select_idx_generic P_((struct folder_browser *dir,
					struct name_vector *entry,
					struct string **buffer));


/* con_map.c */

extern int imap_handle_greeting P_((struct connection_cache *con));

/* common_local.c */

extern int local_selection_stat P_((const char * str));

/* localmbx.c */

extern const struct remote_server * mbx_give_normal_server P_((struct folder_info *folder));

/* remote_mbx.c */

extern int same_remote_account_host_user P_((const struct remote_account *ra1,
					     const struct remote_account *ra2));

extern int remote_account_match P_((const struct remote_account *ra, 
				    const char *username, const char *host));

/* hashmark.c */

extern const unsigned char hashmark_RESERVED_SEPARATOR[];

/* hashmark_browser.c */

extern struct browser_type hashmark_browser;

extern int hashmark_browser_passthru_change P_((struct folder_browser *dir,
						struct hashmark_item *item,
						struct string **dispname));

extern void set_hashmark P_((struct name_vector    *entry,
			     struct hashmark_item *item));

/* hashmark_remote.c */
extern void free_usermap_files P_((void));

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
